require 'game/world/Collision'
require 'game/world/Render'
require 'game/world/Factory'
require 'game/world/Group'
require 'game/world/Map'
require 'game/world/CameraShake'
require 'game/world/Particle'
require 'game/world/HitFrameStop'

World = class('World')
World:include(Collision)
World:include(Render)
World:include(Factory)
World:include(Map)
World:include(CameraShake)
World:include(Particle)
World:include(HitFrameStop)

function World:init(game)
    self.game = game
    self.id = getUID()
    self:collisionInit()
    self:renderInit()
    self:factoryInit()
    self:cameraShakeInit()
    self:particleInit()
    self:hitFrameStopInit()
    love.physics.setMeter(32)
    self.world = self.game.world 
    self.world:setCallbacks(self.collisionOnEnter, self.collisionOnExit)
    self.groups = {}
    self.entities = {}

    self.player = nil
    self:create('Player', 300, 300)
    self.ent_id = nil
    self:mapInit()
    self:addGroup('Projectile')
    self:addGroup('Trail')
    self:addGroup('NumberPopText')
    self:addGroup('InfoUpText')
    self:addCollisionEnter('Projectile', 'Solid', 'collisionSolid')
    self:addCollisionEnter('Projectile', 'Prop', 'collisionProp')
    self:addCollisionEnter('Projectile', 'BSolid', 'collisionBSolid')
    self:addCollisionEnter('Projectile', 'Baby', 'collisionBaby')
    self:addCollisionEnter('Projectile', 'Rogue', 'collisionRogue')
    self:addCollisionEnter('Projectile', 'Monk', 'collisionMonk')
    self:addCollisionEnter('Projectile', 'Ent', 'collisionEnt')
    self:addCollisionEnter('Projectile', 'Entie', 'collisionEntie')
    self:addCollisionEnter('Baby', 'Solid', 'collisionSolid')
    self:addCollisionEnter('Baby', 'BSolid', 'collisionSolid')
    self:addCollisionEnter('Rogue', 'Solid', 'collisionSolid')
    self:addCollisionEnter('Rogue', 'BSolid', 'collisionSolid')
    self:addCollisionEnter('Player', 'Rogue', 'collisionRogueEnter')
    self:addCollisionEnter('Player', 'Entie', 'collisionEntieEnter')
    self:addCollisionExit('Player', 'Rogue', 'collisionRogueExit')
    self:addCollisionExit('Player', 'Entie', 'collisionEntieExit')
    self:addCollisionEnter('Monk', 'Solid', 'collisionSolid')
    self:addCollisionEnter('Monk', 'BSolid', 'collisionSolid')
    self:addCollisionEnter('Player', 'Projectile', 'collisionProjectile')
    self:addCollisionEnter('Player', 'Item', 'collisionItem')
end

function World:update(dt)
    self:hitFrameStopUpdate(dt)
    if self.frame_stopped then return end
    for _, group in ipairs(self.groups) do group:update(dt) end
    table.sort(self.entities, function(e1, e2) return e1.y < e2.y end)
    for _, entity in ipairs(self.entities) do entity:update(dt) end
    self:cameraShakeUpdate(dt)
    self:particleUpdate(dt)
    self:renderUpdate(dt)
    self.world:update(dt)
    self:createPostWorldStep()
    self:removePostWorldStep()
    for _, group in ipairs(self.groups) do group:removePostWorldStep() end
end

function World:draw()
    self:renderAttach()
    self:mapDraw()
    for _, entity in ipairs(self.entities) do 
        if self:entityInCameraView(entity) then
            entity:draw() 
        end
    end
    for _, group in ipairs(self.groups) do group:draw() end
    self:particleDraw()
    for _, entity in ipairs(self.entities) do 
        if self:entityInCameraView(entity) then
            if entity:isInstanceOf(Baby) or entity:isInstanceOf(Rogue) or entity:isInstanceOf(Monk) then
                entity:speakerDraw()
            end
        end
    end
    self:renderDetach()
end

function World:add(entity)
    table.insert(self.entities, entity)
end

function World:remove(id)
    table.remove(self.entities, findIndexByID(self.entities, id))
end

function World:addGroup(group_name)
    table.insert(self.groups, Group(group_name))
end

function World:addToGroup(group_name, entity)
    for _, group in ipairs(self.groups) do
        if group.name == group_name then
            group:add(entity)
            return
        end
    end
end

function World:removeGroup(group_name)
    for i, group in ipairs(self.groups) do
        if group.name == group_name then
            group:destroy()
            table.remove(self.groups, i)
            return
        end
    end
end

function World:removeFromGroup(group_name, id)
    for _, group in ipairs(self.groups) do
        if group.name == group_name then
            group:remove(id)
            return
        end
    end
end

function World:removePostWorldStep()
    for i = #self.entities, 1, -1 do
        if self.entities[i].dead then
            if self.entities[i].class:includes(Timer) then self.entities[i]:timerDestroy() end
            if self.entities[i].class:includes(PhysicsRectangle) then
                self.entities[i].fixture:setUserData(nil)
                self.entities[i].sensor:setUserData(nil)
                self.entities[i].body:destroy()
                self.entities[i].fixture = nil
                self.entities[i].sensor = nil
                self.entities[i].body = nil
            end
            self.entities[i].world = nil
            self:remove(self.entities[i].id)
        end
    end
end

function Group:destroy()
    
end

function World:destroy()
    for _, group in ipairs(self.groups) do
        group:apply(function(entity)
            entity.dead = true
        end)
    end
    for i, group in ipairs(self.groups) do 
        group:removePostWorldStep() 
        group:destroy()
    end
    for i, entity in ipairs(self.entities) do
        entity.dead = true
    end
    self:removePostWorldStep()
    self.entities = nil
    self.groups = nil
    self.player = nil
    self.camera = nil
    self.groups = nil
    self.world = nil
end

function World:mousepressed(x, y, button)

end

function World:mousereleased(x, y, button)

end

function World:keypressed(key)

end

function World:keyreleased(key)

end
